<?php

namespace BexioConnectorByHelvy;

/**
 * WooCommerce to Bexio conversion / parsing
 *
 * @since 0.1.0
 */

defined('ABSPATH') || exit;

class WCToBexio
{
    use Singleton;

    public function __construct() {}

    /**
     * Syncs a WooCommerce Order to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     * @param boolean $force_sync Whether to force the sync or not
     * @return array|WP_Error Bexio order and invoice resources
     */
    public function sync_order($order, $force_sync = false)
    {
        // Pre-sync checks
        if ($order->get_total() == 0 && !get_option(AdminAPISettings::$prefix . 'sync_zero_orders')) {
            $err_msg = "Cannot create invoice for free order (total=0)";
            return new \WP_Error('SYNC_BEXIO_ORDER', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }
        if (!get_option(AdminAPISettings::$prefix . 'sync_orders') && !$force_sync) {
            return ["status" => "SKIPPED", "reason" => "sync_orders disabled"];
        }
        if ($this->is_order_already_synced($order)) return;
        if (!$this->is_order_sync_authorized_based_on_payment_method($order) && !$force_sync) {
            return;
        }

        // Customer sync
        $contacts = $this->find_or_upsert_contact_from_order($order);
        if ($contacts instanceof \WP_Error) {
            return $contacts;
        }
        
        $contact = $contacts["contact"];
        $subcontact = $contacts["subcontact"];

        // Customer's address(es) sync
        $billing_address = $this->upsert_contact_additional_address_from_order($contact['id'], $order, "billing");
        if ($billing_address instanceof \WP_Error) {
            return $billing_address;
        }

        $shipping_address_different_from_billing = ($order->get_shipping_address_1() && ($order->get_billing_address_1() != $order->get_shipping_address_1()))
            || ($order->get_shipping_postcode() && ($order->get_billing_postcode() != $order->get_shipping_postcode()))
            || ($order->get_shipping_city() && ($order->get_billing_city() != $order->get_shipping_city()));
        $shipping_address = $shipping_address_different_from_billing
            ? $this->upsert_contact_additional_address_from_order($contact['id'], $order, "shipping")
            : null; // = Use customers' billing address (same as shipping)
        if ($shipping_address instanceof \WP_Error) {
            return $shipping_address;
        }

        $is_tax_exempt = count($order->get_taxes()) == 0;
        $bexio_line_items = $this->woocommerce_line_items_to_bexio_positions($order, $is_tax_exempt);
        if ($bexio_line_items instanceof \WP_Error) {
            return $bexio_line_items;
        }
        $bexio_fees_lines = $this->woocommerce_fees_to_bexio_positions($order, $is_tax_exempt);
        if ($bexio_fees_lines instanceof \WP_Error) {
            return $bexio_fees_lines;
        }
        $bexio_shipping_lines = $this->woocommerce_shipping_to_bexio_positions($order, $is_tax_exempt);
        if ($bexio_shipping_lines instanceof \WP_Error) {
            return $bexio_shipping_lines;
        }
        $bexio_discount_lines = $this->woocommerce_discount_to_bexio_position($order);
        if ($bexio_discount_lines instanceof \WP_Error) {
            return $bexio_discount_lines;
        }

        // Retrieve Bexio's currency ID or create it if it doesn't exist yet in Bexio
        $bexio_currencies = API::instance()->get_currencies();
        if ($bexio_currencies instanceof \WP_Error) {
            return $bexio_currencies;
        }
        $payment_currency = strtoupper($order->get_currency());
        $currency_id = null;
        foreach ($bexio_currencies as $currency) {
            if ($payment_currency == strtoupper($currency["name"])) {
                $currency_id = $currency["id"];
                break;
            }
        }
        if (!$currency_id) {
            $bexio_currency = API::instance()->create_currency($payment_currency);
            if ($bexio_currency instanceof \WP_Error) {
                return $bexio_currency;
            }
            $currency_id = $bexio_currency["id"];
        }

        $payment_gateway = $order->get_payment_method();
        $bexio_bank_account_id = $this->get_bank_account_from_payment_gateway($payment_gateway, $payment_currency);

        $custom_header_text = $this->parse_header_text($order);
        $custom_footer_text = $this->parse_footer_text($order);

        $use_bexio_taxes = get_option(AdminAPISettings::$prefix . 'use_bexio_products_taxes');
        $prices_include_tax = wc_prices_include_tax() || $use_bexio_taxes;

        // Create order in Bexio
        $bexio_order_body = [
            "title" => $this->generate_bexio_title_from_order($order),
            "is_valid_from" => $this->iso_date_to_bexio_locale_date($order->get_date_created()->__toString()),
            "contact_id" => $contact["id"],
            "contact_sub_id" => $subcontact ? $subcontact["id"] : null,
            "user_id" => get_option(AdminAPISettings::$prefix . "bexio_default_user"),
            "pr_project_id" => null,
            "language_id" => get_option(AdminAPISettings::$prefix . "bexio_default_language"),
            "bank_account_id" => $bexio_bank_account_id,
            "payment_type_id" => null,
            "currency_id" => $currency_id,
            "header" => $custom_header_text,
            "footer" => $custom_footer_text,
            // Always include taxes in invoice (or exempt for shipping outside Switzerland)
            // mwst_type: isTaxExempt ? 2 : 0, // 0=including taxes; 1=excluding taxes; 2=exempt from taxes
            "mwst_type" => 0, // 0=including taxes; exemptions are handled line by line; (1=excluding taxes; 2=exempt from taxes)
            "mwst_is_net" => !$prices_include_tax, // if type=0 -> false=Taxes are included in the total; true=Taxes will be added to the total (relevant only when mwst_type=0)
            "show_position_taxes" => false,
            "api_reference" => $order->get_id(),
            "template_slug" => get_option(AdminAPISettings::$prefix . "bexio_default_document_template")
                ? get_option(AdminAPISettings::$prefix . "bexio_default_document_template")
                : null,
            "positions" => [
                ...$bexio_line_items,
                ...$bexio_fees_lines,
                ...$bexio_shipping_lines,
                ...$bexio_discount_lines,
            ],
        ];

        // Handle custom - non-default - billing/shipping addresses
        if (!empty($billing_address["id"])) {
            $bexio_order_body["contact_address_id"] = $billing_address["id"];
        }
        if ($shipping_address_different_from_billing && !empty($shipping_address["id"])) {
            $bexio_order_body["delivery_address_type"] = 1; // 0 (default) = use billing address; 1 = use another address
            $bexio_order_body["delivery_address_id"] = $shipping_address["id"];
        }

        $bexio_order = API::instance()->create_order($bexio_order_body);
        if ($bexio_order instanceof \WP_Error) {
            return $bexio_order;
        }

        $bexio_invoice = null;
        if (get_option(AdminAPISettings::$prefix . 'auto_create_invoices') && $order->get_total() != 0) {
            $bexio_invoice = API::instance()->create_invoice_from_order($bexio_order["id"]);
            if ($bexio_invoice instanceof \WP_Error) {
                return $bexio_invoice;
            }

            // Set invoice date to match order's date if not already the case
            if ($bexio_order["is_valid_from"] != $bexio_invoice["is_valid_from"]
            || $bexio_order["user_id"] != $bexio_invoice["user_id"]
            || get_option(AdminAPISettings::$prefix . "bexio_default_document_template")
            || $custom_header_text || $custom_footer_text) {
                $updated_invoice = API::instance()->update_invoice($bexio_invoice["id"], [
                    "is_valid_from" => $bexio_order_body["is_valid_from"],
                    "user_id" => get_option(AdminAPISettings::$prefix . "bexio_default_user"),
                    "template_slug" => get_option(AdminAPISettings::$prefix . "bexio_default_document_template")
                        ? get_option(AdminAPISettings::$prefix . "bexio_default_document_template")
                        : null,
                    "header" => $custom_header_text ? $custom_header_text : $bexio_invoice["header"],
                    "footer" => $custom_footer_text ? $custom_footer_text : $bexio_invoice["footer"],
                ]);
                if ($updated_invoice instanceof \WP_Error) {
                    return $updated_invoice;
                }
            }

            // Check if invoice should be issued directly (i.e. not kept in "draft" status)
            if (get_option(AdminAPISettings::$prefix . 'auto_issue_invoices')
                && !$this->keep_invoice_draft_based_on_payment_method($order)) {
                $bexio_issue_invoice = API::instance()->issue_invoice($bexio_invoice["id"]);
                if ($bexio_issue_invoice instanceof \WP_Error) {
                    return $bexio_issue_invoice;
                }
            }
        }

        // Create the delivery for this order
        $bexio_delivery = API::instance()->create_delivery_from_order($bexio_order["id"]);
        if ($bexio_delivery instanceof \WP_Error) {
            return $bexio_delivery;
        }

        // Ensure we are working with the latest order data (patch for when the order is created in "new_order" hook)
        $order = wc_get_order($order->get_id());
        
        // Add metafields to order
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_order_id', $bexio_order["id"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_order_number', $bexio_order["document_nr"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_delivery_id', $bexio_delivery["id"]);
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_delivery_number', $bexio_delivery["document_nr"]);
        if ($bexio_invoice) {
            $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_invoice_id', $bexio_invoice["id"]);
            $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_invoice_number', $bexio_invoice["document_nr"]);
        }
        $order->save_meta_data();

        return ["bexio_order" => $bexio_order, "bexio_invoice" => $bexio_invoice];
    }

    /**
     * Parses the footer text for the invoice based on the settings and the order.
     */
    private function parse_footer_text($order)
    {
        $footer_text = get_option(AdminAPISettings::$prefix . "invoice_footer_text");
        if (!$footer_text) return false; // Return false to indicate no footer text

        $footer_text = str_replace("\n", '<br>', $footer_text);
        $footer_text = str_replace('{woocommerce_order_id}', $order->get_id(), $footer_text);
        $footer_text = str_replace('{woocommerce_order_number}', $order->get_id(), $footer_text);
        $footer_text = str_replace('{woocommerce_order_payment_method}', $order->get_payment_method_title(), $footer_text);
        $footer_text = str_replace('{woocommerce_order_customer_notes}', $order->get_customer_note(), $footer_text);
        $footer_text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $footer_text);
        $footer_text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $footer_text);
        return $footer_text;
    }

    /**
     * Parses the header text for the invoice based on the settings and the order.
     */
    private function parse_header_text($order)
    {
        $header_text = get_option(AdminAPISettings::$prefix . "invoice_header_text");
        if (!$header_text) return false; // Return false to indicate no header text

        $header_text = str_replace("\n", '<br>', $header_text);
        $header_text = str_replace('{woocommerce_order_id}', $order->get_id(), $header_text);
        $header_text = str_replace('{woocommerce_order_number}', $order->get_id(), $header_text);
        $header_text = str_replace('{woocommerce_order_payment_method}', $order->get_payment_method_title(), $header_text);
        $header_text = str_replace('{woocommerce_order_customer_notes}', $order->get_customer_note(), $header_text);
        $header_text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $header_text);
        $header_text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $header_text);
        return trim($header_text);
    }

    /**
     * Syncs a WooCommerce Product to bexio (upsert).
     * 
     * @param Product $product WooCommerce Product resource
     * @param boolean $force_sync Whether to force the sync or not
     * @return array|WP_Error Bexio product or error
     */
    public function sync_product($product, $force_sync = false)
    {
        $is_virtual = $product->is_virtual();
        $is_downloadable = $product->is_downloadable();
        $is_physical = !$is_virtual && !$is_downloadable;

        $is_stock = get_option(AdminAPISettings::$prefix . "sync_stocks") == "BEXIO_TO_WOOCOMMERCE";
        
        // Check if product has a 'bexio_product_id' metafield
        $bexio_product_id = $product->get_meta(AdminAPISettings::$prefix . 'bexio_product_id');

        // Try to find the article in bexio based on its SKU, otherwise based on its "bexio_product_id"
        $bexio_product = null;
        if ($product->get_sku()) {
            $bexio_product = API::instance()->get_article_by_sku($product->get_sku());
        } else if ($bexio_product_id) {
            $bexio_product = API::instance()->get_article_by_id($bexio_product_id);
        }

        $product_name = $product->get_title();
        if ($product instanceof \WC_Product_Variation) {
            $product_name .= ' - ' . wc_get_formatted_variation($product, true, false);
        }

        if ($bexio_product) {
            // If bexio product is found, update it in bexio
            $bexio_product_data = [
                "intern_code" => $product->get_sku(),
                "intern_name" => $product_name,
                "purchase_price" => $product->get_meta('_purchase_price'),
                "sale_price" => $product->get_price(),
                "currency_id" => get_option(AdminAPISettings::$prefix . "bexio_default_currency"),
            ];

            // Optional/conditional fields
            if (get_option(AdminAPISettings::$prefix . "sync_products_description")) $bexio_product_data["intern_description"] = $product->get_description();
            if ($product->get_width()) $bexio_product_data["width"] = round($product->get_width() * 10); // Convert cm to mm and round to integer
            if ($product->get_height()) $bexio_product_data["height"] = round($product->get_height() * 10); // Convert cm to mm and round to integer
            if ($product->get_weight()) $bexio_product_data["weight"] = round($product->get_weight() * 1000); // Convert kg to g and round to integer
            
            $bexio_product = API::instance()->update_article($bexio_product["id"], $bexio_product_data);
        } else {
            // Otherwise, create it in bexio
            $bexio_product = API::instance()->create_article([
                "user_id" => get_option(AdminAPISettings::$prefix . "bexio_default_user"),
                "article_type_id" => $is_physical ? 1 : 2, // 1=physical; 2=service
                "contact_id" => null,
                "deliverer_code" => null,
                "deliverer_name" => null,
                "deliverer_description" => null,
                "intern_code" => $product->get_sku(),
                "intern_name" => $product_name,
                "intern_description" => get_option(AdminAPISettings::$prefix . "sync_products_description") ? $product->get_description() : null,
                "intern_description" => null,
                "purchase_price" => $product->get_meta('_purchase_price'),
                "sale_price" => $product->get_price(),
                "purchase_total" => null,
                "sale_total" => null,
                "currency_id" => get_option(AdminAPISettings::$prefix . "bexio_default_currency"),
                "tax_income_id" => null,
                "tax_expense_id" => null,
                "unit_id" => null,
                "is_stock" => $is_stock,
                "stock_id" => null,
                "stock_place_id" => null,
                "stock_nr" => $product->get_stock_quantity(),
                "stock_min_nr" => 0,
                "width" => $product->get_width() ? round($product->get_width() * 10) : null, // Convert cm to mm and round to integer
                "height" => $product->get_height() ? round($product->get_height() * 10) : null, // Convert cm to mm and round to integer
                "weight" => $product->get_weight() ? round($product->get_weight() * 1000) : null, // Convert kg to g and round to integer
                "volume" => null,
                "remarks" => null,
                "delivery_price" => null,
                "article_group_id" => null,
            ]);
        }
        
        if ($bexio_product instanceof \WP_Error) {
            return $bexio_product;
        }

        // Update product's metafield
        $product->update_meta_data(AdminAPISettings::$prefix . 'bexio_product_id', $bexio_product["id"]);
        $product->save_meta_data();

        return ["bexio_product" => $bexio_product];
    }

    /**
     * Indicates if the payment of the given order should be ignored from sync
     * 
     * @param object $order The WooCommerce order
     * @return boolean True if the payment should be ignored, false otherwise
     */
    public function ignore_payment_sync_based_on_payment_method($order)
    {
        $ignore_payment_methods = get_option(AdminAPISettings::$prefix . 'ignore_payment_sync_for_payment_methods');
        if (empty($ignore_payment_methods) || !is_array($ignore_payment_methods)) return false;
        $order_payment_method = $order->get_payment_method();
        foreach ($ignore_payment_methods as $method) {
            if ($method['paymentMethod'] === $order_payment_method) {
                return true;
            }
        }
        return false;
    }

    /**
     * Indicates if the invoice should stay draft (not issued) based on the payment method
     * 
     * @param object $order The WooCommerce order
     * @return boolean True if the invoice should stay in draft status
     */
    public function keep_invoice_draft_based_on_payment_method($order)
    {
        $ignore_payment_methods = get_option(AdminAPISettings::$prefix . 'ignore_payment_sync_for_payment_methods');
        if (empty($ignore_payment_methods) || !is_array($ignore_payment_methods)) return false;
        $order_payment_method = $order->get_payment_method();
        foreach ($ignore_payment_methods as $method) {
            if ($method['paymentMethod'] === $order_payment_method && $method['keepInvoicesDraft']) {
                return true;
            }
        }
        return false;
    }

    /**
     * Syncs a WooCommerce Order's payment to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     */
    public function sync_order_payment($order, $force_sync = false)
    {
        // Pre-sync checks

        // Check if payments sync is enabled
        if (!get_option(AdminAPISettings::$prefix . 'sync_payments') && !$force_sync) {
            return ["status" => "SKIPPED", "reason" => "sync_payments disabled"];
        }

        // Check if the order is authorized to be synced for this order's payment method
        if ($this->ignore_payment_sync_based_on_payment_method($order)) {
            $err_msg = "Payment sync is disabled for this payment method";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // Check if the order has been paid
        if (!$order->is_paid()) {
            $err_msg = "Order has not been paid yet";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // 1. Retrieve invoice
        $bexio_invoice_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_invoice_id');
        if (!$bexio_invoice_id) {
            $err_msg = "Invoice not yet created for this order";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // 2. Check existing payments (skip if some already exist)
        $existing_payments = API::instance()->get_invoice_payments($bexio_invoice_id);
        if ($existing_payments && is_array($existing_payments) && count($existing_payments)) {
            $err_msg = "Payment(s) already exist in Bexio for this order";
            return new \WP_Error('SYNC_BEXIO_PAYMENT', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // 3. Mark invoice paid by creating the related payment
        $payment_currency = strtoupper($order->get_currency());
        $payment_date = $order->get_date_paid()
            ? $this->iso_date_to_bexio_locale_date($order->get_date_paid())
            : (new \DateTime())->format('Y-m-d');
        
        $bexio_payments = [];

        // Retrieve gift cards (if applicable)
        $giftcards = $order->get_items('gift_card');
        $giftcards_total = 0;
        if ($giftcards) {
            foreach ( $giftcards as $giftcard ) {
                $amount = $giftcard->get_amount();
                if ($amount <= 0) {
                    continue;
                }
                $giftcards_total += $amount;

                $bexio_bank_account_id = $this->get_bank_account_from_payment_gateway("gift_card", $payment_currency);
                $bexio_payments[] = [
                    "date" => $payment_date,
                    "value" => $amount,
                    "bank_account_id" => $bexio_bank_account_id,
                    "payment_service_id" => null,
                ];
            }
        }

        // Retrieve payment currency and gateway
        $payment_gateway = $order->get_payment_method();
        $bexio_bank_account_id = $this->get_bank_account_from_payment_gateway($payment_gateway, $payment_currency);
        $bexio_payments[] = [
            "date" => $payment_date,
            "value" => $order->get_total(),
            "bank_account_id" => $bexio_bank_account_id,
            "payment_service_id" => null,
        ];

        $bexio_payments_res = [];
        foreach ($bexio_payments as $bexio_payment) {
            $bexio_payment_res = API::instance()->create_invoice_payment($bexio_invoice_id, $bexio_payment);
            if ($bexio_payment_res instanceof \WP_Error) {
                return $bexio_payment_res;
            }
            $bexio_payments_res[] = $bexio_payment_res;
        }


        // Ensure we are working with the latest order data (patch for when the order is created in "new_order" hook)
        $order = wc_get_order($order->get_id());

        // Add metafield to order
        $bexio_payments_ids = implode(',', array_column($bexio_payments_res, 'id'));
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_payment_id', $bexio_payments_ids);
        $order->save_meta_data();

        return ["bexio_payment" => $bexio_payments_res];
    }

    /**
     * Syncs a WooCommerce Order's fulfillment to bexio.
     * 
     * @param Order $order WooCommerce Order resource
     */
    public function sync_order_fulfillment($order, $force_sync = false)
    {
        // Check if fulfillments sync is enabled
        if (!get_option(AdminAPISettings::$prefix . 'sync_fulfillments') && !$force_sync) {
            return ["status" => "SKIPPED", "reason" => "sync_fulfillments disabled"];
        }

        $order_status = $order->get_status();
        if ($order_status != 'completed') {
            $err_msg = "Order has not been completed yet (status: " . $order_status . ")";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        $bexio_order_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_order_id');
        if (!$bexio_order_id) {
            $err_msg = "No bexio order associated with this WooCommerce order";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        $bexio_delivery_id = $order->get_meta(AdminAPISettings::$prefix . 'bexio_delivery_id');
        if (!$bexio_delivery_id) {
            $err_msg = "No bexio delivery associated with this WooCommerce order";
            return new \WP_Error('ISSUE_BEXIO_DELIVERY', $err_msg, array(
                'status_code' => 412, // Precondition Failed
            ));
        }

        // 2. Issue the delivery (will mark the order as "Done" in bexio)
        $issued_delivery = API::instance()->issue_delivery($bexio_delivery_id);
        if ($issued_delivery instanceof \WP_Error) {
            return $issued_delivery;
        }

        // Ensure we are working with the latest order data (patch for when the order is created in "new_order" hook)
        $order = wc_get_order($order->get_id());

        // Add metafield to order
        $order->update_meta_data(AdminAPISettings::$prefix . 'bexio_fulfillment_id', $bexio_delivery_id);
        $order->save_meta_data();

        return ["issued_delivery" => $issued_delivery];
    }

    /**
     * Converts an ISO 8601 date into the date format expected by bexio.
     * 
     * @param string $iso_date The ISO 8601 date to convert
     * @return string The Y-m-d formated date expected by bexio
     */
    protected function iso_date_to_bexio_locale_date($iso_date)
    {
        try {
            $date = new \DateTime($iso_date, new \DateTimeZone('UTC'));
            $date->setTimezone(new \DateTimeZone('Europe/Zurich'));
            return $date->format('Y-m-d');
        } catch (\Exception $e) {
            return null;
        }
    }

    /**
     * Get the bexio bank account to be used based on the given payment gateway and payment currency.
     * 
     * @param string $gateway Payment gateway method (usually a "technical" name)
     * @param string $payment_currency Currency used for the order
     * @return array ID of the bexio bank account to use
     */
    protected function get_bank_account_from_payment_gateway($gateway, $payment_currency)
    {
        $bexio_bank_account = get_option(AdminAPISettings::$prefix . "bexio_default_bank_account");
        $custom_payment_methods_to_bank_accounts = get_option(AdminAPISettings::$prefix . "custom_payment_methods_to_bank_accounts");
        if (!empty($custom_payment_methods_to_bank_accounts)) {
            // Check for bank account with currency matching
            $found_in_custom_bank_account_binding_with_currency = null;
            foreach ($custom_payment_methods_to_bank_accounts as $c) {
                if (
                    strtolower($c["paymentMethod"]) === strtolower($gateway) &&
                    strtolower($c["currency"]) === strtolower($payment_currency)
                ) {
                    $found_in_custom_bank_account_binding_with_currency = $c;
                    break;
                }
            }

            if ($found_in_custom_bank_account_binding_with_currency) {
                $bexio_bank_account = $found_in_custom_bank_account_binding_with_currency["bexioBankAccount"];
            } else {
                // Check for bank account with empty currency ("default" mapping)
                $found_in_custom_bank_account_binding = null;
                foreach ($custom_payment_methods_to_bank_accounts as $c) {
                    if (strtolower($c["paymentMethod"]) === strtolower($gateway) && empty($c["currency"])) {
                        $found_in_custom_bank_account_binding = $c;
                        break;
                    }
                }

                if ($found_in_custom_bank_account_binding) {
                    $bexio_bank_account = $found_in_custom_bank_account_binding["bexioBankAccount"];
                }
            }
        }

        return $bexio_bank_account;
    }

    /**
     * Get applicable tax data for shipping in an order.
     *
     * @param WC_Order $order The WooCommerce order.
     * @return array Formatted as WC_Order_Item::get_taxes(), with 'total' and 'subtotal' arrays.
     */
    function get_shipping_taxes_formatted($order) {
        // Ensure the provided object is a valid WooCommerce order
        if (!$order instanceof \WC_Order) {
            return [
                'total' => [],
                'subtotal' => []
            ];
        }

        // Determine the shipping destination from the order
        $destination = [
            'country'  => $order->get_shipping_country(),
            'state'    => $order->get_shipping_state(),
            'postcode' => $order->get_shipping_postcode(),
            'city'     => $order->get_shipping_city(),
        ];

        // Get applicable shipping tax rates using find_shipping_rates()
        $tax_rates = \WC_Tax::find_shipping_rates([
            'country'  => $destination['country'],
            'state'    => $destination['state'],
            'postcode' => $destination['postcode'],
            'city'     => $destination['city'],
        ]);

        // Format the result to match $item->get_taxes() structure
        $taxes = [
            'total' => [],
            'subtotal' => []
        ];

        foreach ($tax_rates as $rate_id => $rate) {
            // Add the tax rate with zero amounts for consistency
            $taxes['total'][$rate_id] = 0;
            $taxes['subtotal'][$rate_id] = 0;
        }

        return $taxes;
    }
	
    /**
     * Retrieves the applicable bexio tax rate based on the given WooCommerce Order Item
     * 
     * @param object $order The WooCommerce order
     * @param object $item The order's item
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @param array/string $bexio_product bexio product linked to the item (if applicable; required only if "use_bexio_taxes" is enabled). Can also be "shipping", "fees" if referring to a shipping line or a fee line.
     * @return string Applicable tax rate (bexio tax ID)
     */
    protected function get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt = false, $bexio_product = null)
    {
        $use_bexio_taxes = get_option(AdminAPISettings::$prefix . 'use_bexio_products_taxes');
        if ($use_bexio_taxes) {
            if ($bexio_product == "shipping" || $bexio_product == "fees") {
                $shipping_tax_rate = get_option(AdminAPISettings::$prefix . 'use_bexio_products_taxes_shipping_tax');
                if ($shipping_tax_rate) {
                    return $shipping_tax_rate;
                } else {
                    return new \WP_Error("MISSING_TAX_MAPPING", "No bexio tax mapping found for shipping item");
                }
            } else if ($bexio_product && is_array($bexio_product) && !empty($bexio_product["tax_id"])) {
                return $bexio_product["tax_id"];
            } else {
                return new \WP_Error("MISSING_TAX_MAPPING", "No bexio tax mapping found for item: " . $item->get_name());
            }
        } 

        $taxes = $item->get_taxes();

        // If taxes are not given because the item is "free shipping", determine the applicable tax rate on-the-fly
        if ($bexio_product == "shipping" && bccomp($item->get_total(), "0.00", 2) === 0
            && (empty($taxes["total"]) || !is_array($taxes["total"]) || !count($taxes["total"]))) {
            $taxes = $this->get_shipping_taxes_formatted($order);
        }

        $applicable_tax = null;
        if (get_option('woocommerce_calc_taxes') !== 'yes' || $is_tax_exempt || empty($taxes["total"]) || !is_array($taxes["total"]) || !count($taxes["total"])) {
            $applicable_tax = get_option(AdminAPISettings::$prefix . 'bexio_default_exempt_tax');
        } else if (is_array($taxes) && count($taxes)) {
            // Item taxes
            $tax_rate_percentage = null;
            
			if (count($taxes['total']) == 1) {
				// Take first tax listed
                $tax_rate_percentage = \WC_Tax::get_rate_percent_value(key($taxes['total']));
			} else {
                // Otherwise, take the first non-zero tax rate
                foreach ($taxes['total'] as $tax_rate_id => $tax_amount) {
                    if ($tax_amount !== "") {
                        $tax_rate_percentage = \WC_Tax::get_rate_percent_value($tax_rate_id);
                        break;
                    }
                }
            }

            // Retrieve applicable bexio tax
            $applicable_tax = null;
            $custom_vat_mapping = get_option(AdminAPISettings::$prefix . 'custom_vat_mapping');
            if (is_array($custom_vat_mapping) && count($custom_vat_mapping)) {
                // Search applicable tax rate including country
                foreach ($custom_vat_mapping as $vat_mapping) {
                    if (
                        bccomp($tax_rate_percentage, round($vat_mapping["vatRate"], 2), 2) === 0 && isset($vat_mapping["country"]) &&
                        strtoupper($vat_mapping["country"]) == strtoupper($order->get_billing_country())
                    ) {
                        $applicable_tax = $vat_mapping["bexioTax"];
                        break;
                    }
                }

                if (!$applicable_tax) {
                    // Search applicable tax rate non-country specific
                    foreach ($custom_vat_mapping as $vat_mapping) {
                        if (bccomp($tax_rate_percentage, round($vat_mapping["vatRate"], 2), 2) === 0 && empty($vat_mapping["country"])) {
                            $applicable_tax = $vat_mapping["bexioTax"];
                            break;
                        }
                    }
                }
            }

            if (!$applicable_tax) {
                return new \WP_Error("MISSING_TAX_MAPPING", "No bexio tax mapping found for " . ($tax_rate_percentage ?? 'N/A') . "%");
            }
        }

        return $applicable_tax;
    }


    /**
     * Converts WooCommerce's order discount to bexio line.
     * 
     * @param object $order The WooCommerce order
     * @return array Bexio positions (containing the shipping line(s))
     */
    protected function woocommerce_discount_to_bexio_position($order)
    {
        if ($order->get_discount_total()) {
            $text = get_option(AdminAPISettings::$prefix . "invoice_discount_line_text");
            $text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $text);
            $text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $text);

            return [
                // [
                //     "type" => "KbPositionCustom",
                //     // "amount" => 1,
                //     // "tax_id" => $applicable_tax,
                //     "text" => $text,
                //     "value" => $order->get_discount_total() * -1,
                //     "is_percentual" => false,
                //     // "discount_in_percent" => discountInPercent,
                //     // "account_id" => $account_id ? $account_id : null,
                // ]
                [
                    "type" => "KbPositionDiscount",
                    "text" => $text,
                    "value" => wc_prices_include_tax() ? $order->get_discount_total() + $order->get_discount_tax() : $order->get_discount_total(),
                    "is_percentual" => false,
                ]
            ];
        } else {
            return [];
        }
    }

    /**
     * Converts WooCommerce's shipping line to bexio positions.
     * 
     * @param object $order The WooCommerce order
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return array Bexio positions (containing the shipping line(s))
     */
    protected function woocommerce_shipping_to_bexio_positions($order, $is_tax_exempt = false)
    {
        $bexio_positions = [];
        foreach ($order->get_items('shipping') as $item) {
            $account_id = get_option(AdminAPISettings::$prefix . 'bexio_default_shipping_income_account');

            $unit_price = wc_prices_include_tax()
                ? wc_format_decimal(((float) $item->get_total() + (float) $item->get_total_tax()) / $item->get_quantity(), wc_get_price_decimals())
                : wc_format_decimal((float) $item->get_total() / $item->get_quantity(), wc_get_price_decimals());

            $text = get_option(AdminAPISettings::$prefix . "invoice_shipping_line_text");
            $text = str_replace('{shipping_line_title}', $item->get_name(), $text);
            $text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $text);
            $text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $text);

            $applicable_tax = $this->get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt, "shipping");
            if ($applicable_tax instanceof \WP_Error) {
				if (bccomp($item->get_total(), "0.00", 2) === 0) {
                    continue; // Skip shipping line with 0 total
                } else {
                    return $applicable_tax;
                }
            }
            $bexio_positions[] = [
                "type" => "KbPositionCustom",
                "amount" => 1,
                "tax_id" => $applicable_tax,
                "text" => $text,
                "unit_price" => $unit_price,
                "account_id" => $account_id ? $account_id : null,
            ];
        }

        return $bexio_positions;
    }

    /**
     * Converts WooCommerce's order line items to bexio positions.
     * 
     * @param object $order The WooCommerce order
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return array Bexio positions
     */
    protected function woocommerce_line_items_to_bexio_positions($order, $is_tax_exempt = false)
    {
        $bexio_positions = [];
        foreach ($order->get_items() as $item_id => $item) {
            $product_variation_id = $item->get_variation_id();
            $product = $product_variation_id
                ? wc_get_product($product_variation_id)
                : wc_get_product($item->get_product_id());

            $requires_shipping = !$product->is_virtual();
            $unit_price = wc_prices_include_tax()
                ? wc_format_decimal(((float) $item->get_subtotal() + (float) $item->get_subtotal_tax()) / $item->get_quantity(), wc_get_price_decimals())
                : wc_format_decimal((float) $item->get_subtotal() / $item->get_quantity(), wc_get_price_decimals());

            $account_id = $requires_shipping
                ? get_option(AdminAPISettings::$prefix . 'bexio_default_product_income_account')
                : get_option(AdminAPISettings::$prefix . 'bexio_default_service_income_account');

            // Try to match item in bexio
            $bexio_product = null;
            if ($product->get_sku()) {
                $bexio_product = API::instance()->get_article_by_sku($product->get_sku());
            }

            $text = $this->generate_line_item_text_from_product($bexio_product, $product);
            $applicable_tax = $this->get_applicable_tax_by_woocommerce_item($order, $item, $is_tax_exempt, $bexio_product);
            if ($applicable_tax instanceof \WP_Error) {
                return $applicable_tax;
            }
            if ($bexio_product) {
                $bexio_position = [
                    "type" => "KbPositionArticle",
                    "article_id" => $bexio_product["id"],
                    "amount" => $item->get_quantity(),
                    "tax_id" => $applicable_tax,
                    "text" => $text,
                    "unit_price" => $unit_price,
                    // account_id: accountId, // -> Deprecated, will use bexio article's account_id
                ];
            } else {
                $bexio_position = [
                    "type" => "KbPositionCustom",
                    "amount" => $item->get_quantity(),
                    "tax_id" => $applicable_tax,
                    "text" => $text,
                    "unit_price" => $unit_price,
                    "account_id" => $account_id ? $account_id : null,
                ];
            }

            $bexio_positions[] = $bexio_position;
        }

        return $bexio_positions;
    }

    /**
     * Converts WooCommerce's order fees to bexio positions.
     * 
     * @param object $order The WooCommerce order
     * @param boolean $is_tax_exempt Whether the order is tax exempt or not
     * @return array Bexio positions
     */
    protected function woocommerce_fees_to_bexio_positions($order, $is_tax_exempt = false)
    {
        $bexio_positions = [];
        foreach ($order->get_fees() as $fee_id => $fee) {

            // Account ID for fees
            $account_id = get_option(AdminAPISettings::$prefix . 'bexio_default_product_income_account');
            $text = $fee->get_name(); // Fee name as the description

            // Fee amount can be contained in the fee amount or the total attributes
            // Usually, amount and total contain the same value. But in some cases, the total excludes taxes while amount includes them.
            $fee_amount = $fee->get_total()
                ? $fee->get_total()
                : $fee->get_amount();

            // Fee amount (including tax if applicable)
            $amount = wc_prices_include_tax()
                ? wc_format_decimal(round(((float) $fee_amount + (float) $fee->get_total_tax()), 5), wc_get_price_decimals())
                : wc_format_decimal(round((float) $fee_amount, 5), wc_get_price_decimals());

            // Get the applicable tax for the fee
            $applicable_tax = $this->get_applicable_tax_by_woocommerce_item($order, $fee, $is_tax_exempt, "fees");
            if ($applicable_tax instanceof \WP_Error) {
                return $applicable_tax;
            }

            // Creating a custom Bexio position for the fee
            $bexio_position = [
                "type" => "KbPositionCustom",
                "amount" => 1, // Fees are usually a single unit
                "tax_id" => $applicable_tax,
                "text" => $text,
                "unit_price" => $amount,
                "account_id" => $account_id ? $account_id : null,
            ];

            $bexio_positions[] = $bexio_position;
        }

        return $bexio_positions;
    }


    /**
     * Generates the line item text of the bexio order/invoice based on the given product
     * 
     * @param object $bexio_product Bexio product corresponding to the given product
     * @param object $wc_product Actual WooCommerce product that is part of the order
     * @return string Generated line item text
     */
    protected function generate_line_item_text_from_product($bexio_product, $wc_product)
    {
        $product_name = $wc_product->get_title();
        if ($wc_product instanceof \WC_Product_Variation) {
            $product_name .= ' - ' . wc_get_formatted_variation($wc_product, true, false);
        }

        $line_item_text = get_option(AdminAPISettings::$prefix . "invoice_line_item_format");
        $line_item_text = str_replace("\n", '<br>', $line_item_text);
        $line_item_text = str_replace('{woocommerce_product_name}', $product_name, $line_item_text);
        $line_item_text = str_replace('{woocommerce_sku}', $wc_product->get_sku(), $line_item_text);
        $line_item_text = str_replace('{woocommerce_product_description}', $wc_product->get_description(), $line_item_text);
        $line_item_text = str_replace('{woocommerce_barcode}', $wc_product->get_meta('_global_unique_id'), $line_item_text);
        $line_item_text = str_replace('{bexio_product_title}', !empty($bexio_product["intern_name"]) ? $bexio_product["intern_name"] : (!empty($wc_product->get_name()) ? $wc_product->get_name() : ""), $line_item_text);
        $line_item_text = str_replace('{bexio_product_description}', !empty($bexio_product["intern_description"]) ? $bexio_product["intern_description"] : "", $line_item_text);
        $line_item_text = str_replace('{bexio_sku}', !empty($bexio_product["intern_code"]) ? $bexio_product["intern_code"] : (!empty($wc_product->get_sku()) ? $wc_product->get_sku() : ""), $line_item_text);
        $line_item_text = preg_replace('/\*{2}([^\*\_]+)\*{2}/', '<b>$1</b>', $line_item_text);
        $line_item_text = preg_replace('/\*{1}([^\*\_]+)\*{1}/', '<i>$1</i>', $line_item_text);

        return $line_item_text;
    }

    /**
     * Generates the bexio order/invoice title based on the given WooCommerce order
     * 
     * @param object $wc_order WooCommerce order
     * @return string Generated line item text
     */
    protected function generate_bexio_title_from_order($wc_order)
    {
        $bexio_title = get_option(AdminAPISettings::$prefix . "invoice_title_format");
        $bexio_title = str_replace('{woocommerce_order_id}', $wc_order->get_id(), $bexio_title);
        $bexio_title = str_replace('{woocommerce_order_number}', $wc_order->get_id(), $bexio_title);
        $bexio_title = str_replace('{woocommerce_order_payment_method}', $wc_order->get_payment_method_title(), $bexio_title);
        $bexio_title = str_replace('{woocommerce_order_title}', $wc_order->get_title(), $bexio_title);
        return $bexio_title;
    }

    /**
     * Checks if the given order is authorized to be synced or not based on the authorized payment methods
     * 
     * @param object $order The WooCommerce order
     * @return boolean
     */
    protected function is_order_sync_authorized_based_on_payment_method($order)
    {
        $payment_methods_option = get_option(AdminAPISettings::$prefix . 'sync_orders_only_for_payment_methods');
        if ($payment_methods_option) {
            $payment_methods_to_sync = array_map('trim', explode("\n", $payment_methods_option));
            $order_payment_method = $order->get_payment_method(); // Payment method slug

            // Check if the order's payment method is in the authorized list
            $found_in_payment_methods_to_sync = in_array($order_payment_method, $payment_methods_to_sync);
            if (!$found_in_payment_methods_to_sync) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks if the given order has already been synced
     * 
     * @param object $order The WooCommerce order
     * @return boolean
     */
    protected function is_order_already_synced($order)
    {
        return $order->get_meta(AdminAPISettings::$prefix . 'bexio_invoice_id')
            || $order->get_meta(AdminAPISettings::$prefix . 'bexio_order_id');
    }

    /**
     * Retrieves the bexio contact (find / upsert) corresponding to the given order's customer
     * 
     * @param object $order The WooCommerce order
     * @return array|WP_Error Bexio contact corresponding to the order's customer
     */
    protected function find_or_upsert_contact_from_order($order)
    {
        $company_name = $order->get_billing_company();

        // Fetch matching bexio contact(s) by email, otherwise by first and last name
        $customer_email = $order->get_billing_email();
        $customer_first_name = $order->get_billing_first_name();
        $customer_last_name = $order->get_billing_last_name();
        $contacts = API::instance()->get_contacts_by_email_or_name($customer_email, $customer_first_name, $customer_last_name);
        if ($contacts instanceof \WP_Error) {
            return $contacts;
        }

        $person_contact = null;
        $company_contact = null; // Both contacts to fill
        if (count($contacts) > 0) {
            foreach ($contacts as $contact) {
                if ($contact['contact_type_id'] == 2) {
                    $person_contact = $contact;
                }
            }
            foreach ($contacts as $contact) {
                if ($contact['contact_type_id'] == 1) {
                    $company_contact = $contact;
                }
            }
        }

        // If the person contact is found, update it; otherwise, create it
        $person_contact = $this->upsert_customer($order, $person_contact);
        if ($person_contact instanceof \WP_Error) {
            return $person_contact;
        }

        // Handle company contact / information
        if ($company_name) {
            // If available, update the company information and link the company to the order, not the physical person
            if (!$company_contact) {
                // Search contacts owning this person (i.e. their company)
                $company_relations = API::instance()->search_contact_relations([["field" => "contact_sub_id", "value" => $person_contact['id'], "criteria" => "="]]);
                if ($company_relations instanceof \WP_Error) {
                    return $company_relations;
                }
                if (is_array($company_relations) && count($company_relations)) {
                    foreach ($company_relations as $company_relation) {
                        $contact = API::instance()->get_contact_by_id($company_relation['contact_id']);
                        if ($contact instanceof \WP_Error) {
                            return $contact;
                        }
                        if (strtolower($contact['name_1']) == strtolower($company_name)) {
                            $company_contact = $contact;
                            break;
                        }
                    }
                }

                // Otherwise, search companyContact directly by name & city (not name only to prevent match on homonyms)
                if (!$company_contact) {
                    $company_contacts = API::instance()->get_contacts_companies_by_name_and_postcode($company_name, $order->get_billing_postcode());
                    if ($company_contacts instanceof \WP_Error) {
                        return $company_contacts;
                    }
                    if (count($company_contacts)) {
                        $company_contact = $company_contacts[0];
                        // In this case, create the relation between the person and the company in bexio
                        $contact_relation = API::instance()->create_contact_relation($company_contact['id'], $person_contact['id']);
                        if ($contact_relation instanceof \WP_Error) {
                            return $contact_relation;
                        }
                    }
                }
            }

            // If found, update company's contact information; otherwise, create the company contact and the relation with the existing customer
            $company_contact = $this->upsert_customer($order, $company_contact, $person_contact['id']);
            if ($company_contact instanceof \WP_Error) {
                return $company_contact;
            }
        }

        if ($company_contact) {
            // The customer is a company, if a "person contact" exist for this company, use it as subcontact
            return [ "contact" => $company_contact, "subcontact" => $person_contact ];
        } else {
            // The customer is a private person
            return [ "contact" => $person_contact, "subcontact" => null ];
        }
    }

    /**
     * Upsert the customer of the given WooCommerce order.
     * If the bexio contact is empty, the contact is created in bexio.
     * Otherwise, the given bexio contact is updated based on the order's customer information.
     * 
     * @param object $order The WooCommerce order
     * @param object $bexio_contact The contact corresponding to the order's customer, if already found in bexio contacts
     * @param string $is_company_owning_contact_id Provide a value if the given contact is a company that owns the given contact ID (for binding)
     * @return array Upserted bexio contact
     */
    protected function upsert_customer($order, $bexio_contact, $is_company_owning_contact_id = null)
    {
        $company_name = $order->get_billing_company();

        $billing_address_line = $order->get_billing_address_1();
        if ($order->get_billing_address_2()) {
            $billing_address_line .= "\n" . $order->get_billing_address_2();
        }

        if ($bexio_contact) {
            $contact_data = []; // Potential customer's data to update

            // Update contact's default (billing) address if it has changed
            if (!$bexio_contact['address'] || ($billing_address_line && $bexio_contact['address'] != $billing_address_line)) {
                $billing_address_country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
                if ($billing_address_country instanceof \WP_Error) {
                    return $billing_address_country;
                }

                $contact_data["address"] = $billing_address_line;
                $contact_data["postcode"] = $order->get_billing_postcode();
                $contact_data["city"] = $order->get_billing_city();
                $contact_data["country_id"] = $billing_address_country['id'];
            }

            // Update phone only if none is set in bexio (fix phone for companies, mobile phone for persons)
            if ($order->get_billing_phone()) {
                if ($company_name && !$bexio_contact["phone_fixed"]) {
                    $contact_data["phone_fixed"] = $order->get_billing_phone();
                } else if (!$company_name && !$bexio_contact["phone_mobile"]) {
                    $contact_data["phone_mobile"] = $order->get_billing_phone();
                }
            }

            // Update Bexio customer (if some data has changed)
            if (count($contact_data) > 0) {
                $bexio_contact = API::instance()->update_contact_by_id($bexio_contact['id'], $bexio_contact);
                if ($bexio_contact instanceof \WP_Error) {
                    return $bexio_contact;
                }
            }
        } else {
            $billing_address_country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
            if ($billing_address_country instanceof \WP_Error) {
                return $billing_address_country;
            }

            if ($is_company_owning_contact_id) {
                // Create a company
                $contact_object = [
                    "contact_type_id" => 1,
                    "name_1" => $company_name,
                    "address" => $billing_address_line,
                    "postcode" => $order->get_billing_postcode(),
                    "city" => $order->get_billing_city(),
                    "country_id" => $billing_address_country['id'],
                    "mail" => $order->get_billing_email(),
                    "phone_fixed" => $order->get_billing_phone(),
                    "phone_fixed_second" => null,
                    "phone_mobile" => null,
                    "fax" => null,
                    "url" => null,
                    "skype_name" => null,
                    "remarks" => null,
                    "language_id" => null,
                    "contact_group_ids" => get_option(AdminAPISettings::$prefix . 'bexio_default_contacts_group'), // If list, coma-separated string, e.g. "1,2"
                    "user_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                    "owner_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                ];
                $bexio_contact = API::instance()->create_contact($contact_object);
                if (!empty($bexio_contact['id'])) {
                    API::instance()->create_contact_relation($bexio_contact['id'], $is_company_owning_contact_id);
                }
            } else {
                // Create a physical person
                $bexio_contact = API::instance()->create_contact([
                    "contact_type_id" => 2,
                    "name_1" => $order->get_billing_last_name(), // Empty string if no name provided (prevent bexio error)
                    "name_2" => $order->get_billing_first_name(),
                    "salutation_id" => null,
                    "salutation_form" => null,
                    "titel_id" => null,
                    "birthday" => null,
                    "address" => $billing_address_line,
                    "postcode" => $order->get_billing_postcode(),
                    "city" => $order->get_billing_city(),
                    "country_id" => $billing_address_country['id'],
                    "mail" => $order->get_billing_email(),
                    "phone_fixed" => null,
                    "phone_fixed_second" => null,
                    "phone_mobile" => $order->get_billing_phone(),
                    "fax" => null,
                    "url" => null,
                    "skype_name" => null,
                    "remarks" => null,
                    "language_id" => null,
                    "contact_group_ids" => get_option(AdminAPISettings::$prefix . 'bexio_default_contacts_group'), // If list, coma-separated string, e.g. "1,2"
                    "contact_branch_ids" => null,
                    "user_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                    "owner_id" => get_option(AdminAPISettings::$prefix . 'bexio_default_user'),
                ]);
            }
        }

        return $bexio_contact;
    }

    /**
     * Converts an order's address (billing or shipping) to bexio address format
     * 
     * @param object $order The WooCommerce order resource
     * @param string $address_type The type of address to convert (billing or shipping, defaults to billing)
     * @return array Bexio-structured address
     */
    protected function order_address_to_bexio_address($order, $address_type = "billing")
    {
        if ($address_type == "shipping") {
            $address = $order->get_shipping_address_1();
            if ($order->get_shipping_address_2()) {
                $address .= "\n" . $order->get_shipping_address_2();
            }

            $country = API::instance()->find_or_create_country_by_country_code($order->get_shipping_country());
            if ($country instanceof \WP_Error) {
                return $country;
            }

            $name = $order->get_shipping_company()
                ? $order->get_shipping_company() . "\n" . $order->get_formatted_shipping_full_name()
                : $order->get_formatted_shipping_full_name();

            return [
                "name" => $name,
                "address" => $address,
                "postcode" => $order->get_shipping_postcode() ? $order->get_shipping_postcode() : "0000",
                "city" => $order->get_shipping_city(),
                "country_id" => $country['id'],
            ];
        } else { // Billing
            $address = $order->get_billing_address_1();
            if ($order->get_billing_address_2()) {
                $address .= "\n" . $order->get_billing_address_2();
            }
            $country = API::instance()->find_or_create_country_by_country_code($order->get_billing_country());
            if ($country instanceof \WP_Error) {
                return $country;
            }

            $name = $order->get_billing_company()
                ? $order->get_billing_company() . "\n" . $order->get_formatted_billing_full_name()
                : $order->get_formatted_billing_full_name();

            return [
                "name" => $name,
                "address" => $address,
                "postcode" => $order->get_billing_postcode() ? $order->get_billing_postcode() : "0000",
                "city" => $order->get_billing_city(),
                "country_id" => $country['id'],
            ];
        }
    }

    /**
     * Updates the bexio contact with the given order's address
     * 
     * @param string $contact_id Bexio contact ID
     * @param object $order The WooCommerce order resource
     * @param string $address_type Whether the address is "billing" or "shipping"
     * @return array Created or updated additional adress
     */
    protected function upsert_contact_additional_address_from_order($contact_id, $order, $address_type)
    {
        $address_subject = "WooCommerce " . $address_type . " address";
        $additional_addresses = API::instance()->get_contact_additional_addresses_by_subject($contact_id, $address_subject);
        $existing_address_id = count($additional_addresses) ? $additional_addresses[0]['id'] : 0;

        $bexio_address_obj = $this->order_address_to_bexio_address($order, $address_type);
        if ($bexio_address_obj instanceof \WP_Error) {
            return $bexio_address_obj;
        }

        $bexio_address_obj['subject'] = $address_subject;
        $bexio_address_obj['description'] = $address_subject . " of the customer - Last update " . (new \DateTime())->format('c');

        return $existing_address_id
            ? API::instance()->update_contact_additional_address($contact_id, $existing_address_id, $bexio_address_obj)
            : API::instance()->create_contact_additional_addresses($contact_id, $bexio_address_obj);
    }

    /**
     * Cloning is forbidden.
     */
    public function __clone()
    {
        _doing_it_wrong(__FUNCTION__, __('Cheatin&#8217; huh?'), BEXIO_CONNECTOR_BY_HELVY_VERSION);
    }
}
